/*
 * Part of the GaBIEn project.
 * (Graphics and Basic Input Engine)
 *
 *        DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *                    Version 2, December 2004 
 *
 * Modified by gamemanj <asdd2808@gmail.com> to remove expletives
 * Original copyright (C) 2004 Sam Hocevar <sam@hocevar.net> 
 *
 * Everyone is permitted to copy and distribute verbatim or modified 
 * copies of this license document, and changing it is allowed as long 
 * as the name is changed. 
 *
 *            DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION 
 *
 *  0. You just DO WHATEVER YOU WANT TO.
 */


package gabien;

import java.io.InputStream;
import java.io.OutputStream;

/*
 * This instance should be the entry point to the program.
 * It should call the static method gabienapp.Application.gabienmain(),
 * which runs the application.
 * On UI-thread oriented platforms,
 * please run this on a different thread
 * (All drawing primitives from platforms I've seen
 *  can be accessed from any thread,probably because games
 *  like to use their own schedulers)
 */

public interface IGaBIEn {
	public double getTime();
	public double timeDelta(boolean reset);	

	// Gets a resource from the application binary.
    // Failing that, returns null.
	public InputStream getResource(String resource);
	
	// Runs getResource, then looks for a file in the data directory.
	// Failing that, returns null.
	public InputStream getFile(String resource);
	
	// Creates or overwrites a file in the data directory.
	// Failing that, returns null.
    public OutputStream getOutFile(String resource);
    
	//Means that only one IGrInDriver can be active at a time.
	//Typically,the last will be active,and all others will be ignored.	
	public boolean singleWindowApp();
	public ISoundDriver getSound();//This is a singleton.
	public void ensureQuit();
	//On SingleWindowApp-style platforms,where windowing doesn't exist,ignore windowspecs.
	public IGrInDriver makeGrIn(String name,int w,int h,WindowSpecs windowspecs);
	public WindowSpecs defaultWindowSpecs(String name, int w, int h);
}
