/*
 * Part of the JavaChan project.
 * (Java ikaChan)
 *
 *        DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *                    Version 2, December 2004 
 *
 * Modified by gamemanj <asdd2808@gmail.com> to remove expletives
 * Original copyright (C) 2004 Sam Hocevar <sam@hocevar.net> 
 *
 * Everyone is permitted to copy and distribute verbatim or modified 
 * copies of this license document, and changing it is allowed as long 
 * as the name is changed. 
 *
 *            DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION 
 *
 *  0. You just DO WHATEVER YOU WANT TO.
 */
package javachan;

import gabien.ui.Rect;

import java.util.LinkedList;
import java.util.List;

import javachan.PhysicsEngine;
import javachan.blocks.BasicBlockType;
import javachan.bosses.BasicBoss;

/**
 * "UNLESS someone like you cares a whole awful lot,nothing will ever change."
 * Makes about as much sense to people who like NXEngine for Ikachan, as it does
 * to anything else.
 * 
 * @author matt
 */
public class PhysicsEngine {

    public PhysicsEngine() {
    }

    // Internal interface because this isn't using Java 8 so no lambdas.
    private interface IPhysicsCollider {
        public boolean isColliding(Rect r);
    }

    // Note that it's the PC collisions that create events.This is important.
    // Furthermore,the whole Capacitor thing is one massive special case-
    // -1:Not needed,0:Left,1:Right,2:Up,3:Down

    private int pushOutObjectFromObject(IPhysicsObject po,final Rect r) {
        if (r.intersects(po.getRectangle())) {
            return pushOutObjectFromPhysicsCollider(po, new IPhysicsCollider() {

                @Override
                public boolean isColliding(Rect r2) {
                    return r.intersects(r2);
                }
                
            });
        }
        return -1;
    }

    public int utilClamp(int n, int min, int max) {
        if (n < min) {
            n = min;
        }
        if (n > max) {
            n = max;
        }
        return n;
    }

    public int runPushOutOfMapPass(IPhysicsObject op, GameState gs, boolean soft) {
        GameStateMap map = gs.map;
        int cx = (int) op.getX();
        int cl = cx - 1;
        int cr = cx + 1;
        cl = utilClamp(cl, 0, map.width - 1);
        cr = utilClamp(cr, 0, map.width - 1);
        int cy = (int) op.getY();
        int cu = cy - 1;
        int cd = cy + 1;
        cu = utilClamp(cu, 0, map.height - 1);
        cd = utilClamp(cd, 0, map.height - 1);
        List<IntegerPoint> blocks = new LinkedList<IntegerPoint>();
        for (int px = cl; px <= cr; px++) {
            for (int py = cu; py <= cd; py++) {
                int md = map.data[px + (py * map.width)];
                BasicBlockType bbt = map.blockTypes[md / 16];
                if (bbt != null) {
                    Rect br = bbt.getCollision(px, py, md % 16);
                    if (br.intersects(op.getRectangle())) {
                        if (bbt.collisionStops() || soft) {
                            IntegerPoint ip = new IntegerPoint();
                            ip.X = px;
                            ip.Y = py;
                            blocks.add(ip);
                        }
                    }
                }
            }
        }
        if (blocks.size() > 0) {
            if (soft) {
                for (IntegerPoint ip : blocks) {
                    int md = map.data[ip.X + (ip.Y * map.width)];
                    BasicBlockType bbt = map.blockTypes[md / 16];
                    bbt.handleSoftCollision(ip.X, ip.Y, gs, op);
                }
            } else {
                int d = pushOutObjectFromWorld(op, map);
                for (IntegerPoint ip : blocks) {
                    int md = map.data[ip.X + (ip.Y * map.width)];
                    BasicBlockType bbt = map.blockTypes[md / 16];
                    bbt.handleHardCollision(ip.X, ip.Y, gs, op, d);
                }
                return d;
            }
        }
        return -1;
    }

    private int pushOutObjectFromWorld(final IPhysicsObject op, final GameStateMap map) {
        return pushOutObjectFromPhysicsCollider(op, new IPhysicsCollider() {

            @Override
            public boolean isColliding(Rect r) {
                int cx = (int) op.getX();
                int cl = cx - 1;
                int cr = cx + 1;
                cl = utilClamp(cl, 0, map.width - 1);
                cr = utilClamp(cr, 0, map.width - 1);
                int cy = (int) op.getY();
                int cu = cy - 1;
                int cd = cy + 1;
                cu = utilClamp(cu, 0, map.height - 1);
                cd = utilClamp(cd, 0, map.height - 1);
                for (int px = cl; px <= cr; px++) {
                    for (int py = cu; py <= cd; py++) {
                        int md = map.data[px + (py * map.width)];
                        BasicBlockType bbt = map.blockTypes[md / 16];
                        if (bbt != null) {
                            Rect br = bbt.getCollision(px, py, md % 16);
                            if (br.intersects(op.getRectangle())) {
                                if (bbt.collisionStops()) {
                                    return true;
                                }
                            }
                        }
                    }
                }
                return false;
            }

        });
    }

    private int pushOutObjectFromPhysicsCollider(IPhysicsObject op, IPhysicsCollider iPhysicsCollider) {
        double oX = op.getX(), oY = op.getY();
        int rI = 0, rJ = 0, rD = Integer.MAX_VALUE;

        for (int i = -8; i <= 8; i++)
            for (int j = -8; j <= 8; j++) {
                op.setX(oX + (i / 16.0d));
                op.setY(oY + (j / 16.0d));
                if (!iPhysicsCollider.isColliding(op.getRectangle())) {
                    int d = (i < 0 ? -i : i);
                    d += (j < 0 ? -j : j);
                    if (rD > d) {
                        rD = d;
                        rI = i;
                        rJ = j;
                    }
                }
            }
        int[] iArray = new int[] { 0, 2, 1, 0, -1, 1, 0, 3, 1 };
        int xDir = 1;
        int yDir = 1;
        if (rI < 0)
            xDir = 0;
        if (rI > 0)
            xDir = 2;
        if (rJ < 0)
            yDir = 0;
        if (rJ > 0)
            yDir = 2;
        op.setX(oX + (rI / 16.0d));
        op.setY(oY + (rJ / 16.0d));
        return iArray[xDir + (yDir * 3)];
    }

    public int runPushOutOfObjectPass(IPhysicsObject op, GameState gs, double Time, IEventCallback iec) {
        final GameStateNPChar map = gs.npcs;
        for (int px = 0; px < map.npcTable.length; px++) {
            final GameStateNPChar.NPCCharacter npc=map.npcTable[px];
            if (npc == op) {
                continue;
            }
            if (!npc.exists) {
                continue;
            }
            // collisiontype:0:NULL 1:BLOCK 2:ENEMY 3:EVENT
            int ctype = npc.collisionType;
            boolean Colliding = npc.getRectangle().intersects(op.getRectangle());
            if (Colliding) {
                if ((ctype == 0) || (ctype == 3)) {
                    if (op == gs) {
                        npc.handleIkachanCollision(gs, iec, Time);
                    }
                }
                if ((ctype == 2) || (ctype == 3)) {
                    int r = pushOutObjectFromObject(op, npc.getRectangle());
                    if (gs == op) {
                        if (ctype == 2) {
                            map.npcTypes[npc.entityType].ikachanTouchedEnemy(npc, gs, r, iec);
                        }
                    }
                    if ((r == 3) || (r == 2)) {
                        op.setVY(0);
                    }
                    if ((r == 1) || (r == 0)) {
                        op.setVX(0);
                    }
                    return r;
                }
            }
        }
        return -1;
    }

    private boolean[] collisionPass(IPhysicsObject op, GameState gs, double Time, IEventCallback iec, boolean lastpass) {
        boolean[] ret = new boolean[4];
        boolean IsIkachan = (op == gs);

        int r = 0;
        if (IsIkachan) {
            r = runPushOutOfObjectPass(op, gs, Time, iec);
            if (r != -1) {
                ret[r] = true;
            }
        }
        r = runPushOutOfMapPass(op, gs, false);
        if (r != -1) {
            ret[r] = true;
        }
        if (lastpass) {
            runPushOutOfMapPass(op, gs, true);
            if (IsIkachan) {
                for (BasicBoss boss : gs.bosses) {
                    if (boss.isAlive(gs)) {
                        int hitironhead = pushOutObjectFromObject(op, boss.getRectangle());
                        if (hitironhead != -1) {
                            ret[hitironhead] = true;
                            boss.onHit(gs, hitironhead, iec, Time);
                        }
                    }
                }
            }
        }
        return ret;
    }

    public boolean[] doUpdate(IEventCallback ev, IPhysicsObject ipo, GameState gs, double InGameTime) {
        boolean[] b = collisionPass(ipo, gs, InGameTime, ev, false);
        boolean[] aa = collisionPass(ipo, gs, InGameTime, ev, true);
        aa[0] = aa[0] || b[0];
        aa[1] = aa[1] || b[1];
        aa[2] = aa[2] || b[2];
        aa[3] = aa[3] || b[3];
        cancelVelocityInDirections(aa, ipo);
        return aa;
    }

    // 0:Left,1:Right,2:Up,3:Down
    // Note that these directions are the directions that the player was pushed
    // in to move them out of a block.
    private void cancelVelocityInDirections(boolean[] dirs, IPhysicsObject op) {
        if (dirs[0] && (op.getVX() > 0))
            op.setVX(0);
        if (dirs[1] && (op.getVX() < 0))
            op.setVX(0);
        if (dirs[2] && (op.getVY() > 0))
            op.setVY(0);
        if (dirs[3] && (op.getVY() < 0))
            op.setVY(0);
    }
}
